@@ -5,8 +5,10 @@ import os
 import sys
 
 from ICCProfile import ICCProfile
+from argyll_cgats import extract_device_gray_primaries
 from config import (enc, get_data_path, get_verified_path, getcfg, hascfg,
 					profile_ext, setcfg)
+from debughelpers import Error
 from log import log, safe_print
 from meta import name as appname
 from options import debug
@@ -15,13 +17,15 @@ from util_io import Files
 from util_os import waccess
 from util_str import safe_str
 from worker import Error, FilteredStream, LineBufferedStream, show_result_dialog
+import CGATS
 import ICCProfile as ICCP
 import colormath
 import config
 import localization as lang
 import worker
-from wxwindows import BaseApp, BaseFrame, FileDrop, InfoDialog, wx
+from wxwindows import BaseApp, BaseFrame, ConfirmDialog, FileDrop, InfoDialog, wx
 from wxfixes import TempXmlResource
+from wxLUT3DFrame import LUT3DFrame
 import floatspin
 import xh_floatspin
 import xh_bitmapctrls
@@ -31,7 +35,12 @@ from wx import xrc
 
 class SynthICCFrame(BaseFrame):
 
-	""" 3D LUT creation window """
+	""" Synthetic ICC creation window """
+
+	# Shared methods from 3D LUT UI
+	for lut3d_ivar_name, lut3d_ivar in LUT3DFrame.__dict__.iteritems():
+		if lut3d_ivar_name.startswith("lut3d_"):
+			locals()[lut3d_ivar_name] = lut3d_ivar
 	
 	def __init__(self, parent=None):
 		self.res = TempXmlResource(get_data_path(os.path.join("xrc", 
@@ -64,12 +73,15 @@ class SynthICCFrame(BaseFrame):
 		presets = [""] + sorted(colormath.rgb_spaces.keys())
 		self.preset_ctrl.SetItems(presets)
 
+		self.set_default_cat()
+
 		self.worker = worker.Worker(self)
 
 		# Drop targets
 		self.droptarget = FileDrop(self)
 		self.droptarget.drophandlers = {".icc": self.drop_handler,
-										".icm": self.drop_handler}
+										".icm": self.drop_handler,
+										".ti3": self.drop_handler}
 		self.panel.SetDropTarget(self.droptarget)
 
 		# Bind event handlers
@@ -86,10 +98,13 @@ class SynthICCFrame(BaseFrame):
 						  getattr(self, "%s_%s" % (color, component)))
 		self.black_point_cb.Bind(wx.EVT_CHECKBOX,
 								 self.black_point_enable_handler)
+		self.chromatic_adaptation_btn.Bind(wx.EVT_BUTTON,
+										   self.chromatic_adaptation_btn_handler)
 		self.trc_ctrl.Bind(wx.EVT_CHOICE, self.trc_ctrl_handler)
 		self.trc_gamma_ctrl.Bind(wx.EVT_COMBOBOX, self.trc_gamma_ctrl_handler)
 		self.trc_gamma_ctrl.Bind(wx.EVT_KILL_FOCUS, self.trc_gamma_ctrl_handler)
 		self.trc_gamma_type_ctrl.Bind(wx.EVT_CHOICE, self.trc_gamma_type_ctrl_handler)
+		self.lut3d_bind_hdr_trc_handlers()
 		self.black_output_offset_ctrl.Bind(wx.EVT_SLIDER,
 										   self.black_output_offset_ctrl_handler)
 		self.black_output_offset_intctrl.Bind(wx.EVT_TEXT,
@@ -105,7 +120,6 @@ class SynthICCFrame(BaseFrame):
 		self.save_as_btn.Bind(wx.EVT_BUTTON, self.save_as_btn_handler)
 
 		self.save_as_btn.SetDefault()
-		self.save_as_btn.Disable()
 		
 		self.setup_language()
 		
@@ -146,13 +160,13 @@ class SynthICCFrame(BaseFrame):
 			config.writecfg(module="synthprofile",
 							options=("synthprofile.", "last_icc_path",
 									 "position.synthiccframe",
-									 "size.synthiccframe"))
+									 "size.synthiccframe",
+									 "3dlut.hdr_"))
 		if event:
-			if sys.platform == "win32" and sys.getwindowsversion() >= (6, ):
-				# For some reason, double buffering prevents SynthICC from
-				# exiting cleanly under Windows 10
-				self.panel.SetDoubleBuffered(False)
-			event.Skip()
+			# Hide first (looks nicer)
+			self.Hide()
+			# Need to use CallAfter to prevent hang under Windows if minimized
+			wx.CallAfter(self.Destroy)
 	
 	def black_luminance_ctrl_handler(self, event):
 		v = self.black_luminance_ctrl.GetValue()
@@ -162,8 +176,8 @@ class SynthICCFrame(BaseFrame):
 				wx.Bell()
 			v = white_Y * .9
 		if event:
-			min_Y = (1 / 65535.0) * 100
-			increment = (1 / 65535.0) * white_Y
+			min_Y = 0.000001 #(1 / 65535.0) * 100
+			increment = 0.000001 #(1 / 65535.0) * white_Y
 			if increment < min_Y:
 				increment = min_Y * (white_Y / 100.0)
 			min_inc = 1.0 / (10.0 ** self.black_luminance_ctrl.GetDigits())
@@ -190,11 +204,11 @@ class SynthICCFrame(BaseFrame):
 		if (v != old and (old == 0 or v == 0)) or event is True:
 			self.Freeze()
 			i = self.trc_ctrl.GetSelection()
-			self.trc_gamma_type_ctrl.Show(i in (0, 4) and bool(v))
+			self.trc_gamma_type_ctrl.Show(i in (0, 5) and bool(v))
 			if not v:
 				self.bpc_ctrl.SetValue(False)
 			self.bpc_ctrl.Enable(bool(v))
-			black_output_offset_show = (i in (0, 4, 6, 7) and bool(v))
+			black_output_offset_show = (i in (0, 5, 7, 8) and bool(v))
 			self.black_output_offset_label.Show(black_output_offset_show)
 			self.black_output_offset_ctrl.Show(black_output_offset_show)
 			self.black_output_offset_intctrl.Show(black_output_offset_show)
@@ -233,8 +247,7 @@ class SynthICCFrame(BaseFrame):
 					for i in xrange(3):
 						getattr(self, "black_%s" % "XYZ"[i]).SetValue(0)
 					break
-		for i, v in enumerate(colormath.XYZ2xyY(*XYZ)[:2]):
-			getattr(self, "black_%s" % "xy"[i]).SetValue(v)
+		self.parse_XYZ("black")
 
 	def black_xy_ctrl_handler(self, event):
 		# Black Y scaled to 0..1 range
@@ -253,24 +266,104 @@ class SynthICCFrame(BaseFrame):
 	
 	def blue_xy_ctrl_handler(self, event):
 		self.parse_xy("blue")
+
+	def chromatic_adaptation_btn_handler(self, event):
+		scale = getcfg("app.dpi") / config.get_default_dpi()
+		if scale < 1:
+			scale = 1
+		dlg = ConfirmDialog(self, title=lang.getstr("chromatic_adaptation"),
+							msg=lang.getstr("whitepoint.xy"),
+							ok=lang.getstr("apply"),
+							cancel=lang.getstr("cancel"))
+		sizer = wx.BoxSizer(wx.HORIZONTAL)
+		dlg.sizer3.Add(sizer, 0, flag=wx.TOP | wx.ALIGN_LEFT,
+					   border=8)
+		x_ctrl = floatspin.FloatSpin(dlg, -1, size=(75 * scale, -1),
+									 min_val=0.0001, max_val=1,
+									 increment=0.0001,
+									 value=self.white_x.GetValue())
+		sizer.Add(x_ctrl, 0, flag=wx.RIGHT | wx.ALIGN_CENTER_VERTICAL,
+				  border=4)
+		sizer.Add(wx.StaticText(dlg, -1, u"x"), 0,
+								flag=wx.RIGHT | wx.ALIGN_CENTER_VERTICAL,
+								border=12)
+		y_ctrl = floatspin.FloatSpin(dlg, -1, size=(75 * scale, -1),
+									 min_val=0.0001, max_val=1,
+									 increment=0.0001,
+									 value=self.white_y.GetValue())
+		sizer.Add(y_ctrl, 0, flag=wx.RIGHT | wx.ALIGN_CENTER_VERTICAL,
+				  border=4)
+		sizer.Add(wx.StaticText(dlg, -1, u"y"), 0,
+								flag=wx.RIGHT | wx.ALIGN_CENTER_VERTICAL,
+								border=12)
+		dlg.sizer3.Add(wx.StaticText(dlg, -1,
+									 lang.getstr("chromatic_adaptation_transform")),
+					   0, flag=wx.TOP | wx.ALIGN_LEFT, border=12)
+		if getcfg("show_advanced_options"):
+			# Only offer those that seem suitable in an ICC workflow
+			# (i.e. blue primary not too far from default Bradford to
+			# prevent 'blue turns purple' problems)
+			cat_choices = ['Bradford', 'CAT02BS']
+		else:
+			cat_choices = ["Bradford"]
+		cat_choices_ab = OrderedDict(get_mapping(((k, k) for k in
+												  colormath.cat_matrices),
+												 cat_choices))
+		cat_choices_ba = OrderedDict((v, k) for k, v in cat_choices_ab.iteritems())
+		cat_ctrl = wx.Choice(dlg, -1, choices=cat_choices_ab.values())
+		cat_ctrl.SetStringSelection(cat_choices_ab[self.cat])
+		dlg.sizer3.Add(cat_ctrl, 0, flag=wx.TOP | wx.ALIGN_LEFT, border=8)
+		dlg.sizer0.SetSizeHints(dlg)
+		dlg.sizer0.Layout()
+		result = dlg.ShowModal()
+		x = x_ctrl.GetValue()
+		y = y_ctrl.GetValue()
+		cat = cat_choices_ba[cat_ctrl.GetStringSelection()]
+		dlg.Destroy()
+		if result != wx.ID_OK:
+			return
+		wp_src = [getattr(self, "white_" + component).GetValue()
+				  for component in "XYZ"]
+		wp_tgt = colormath.xyY2XYZ(x, y)
+		self.cat = cat
+		for color in ("red", "green", "blue", "white", "black"):
+			ctrls = [getattr(self, "%s_%s" % (color, component))
+					 for component in "XYZ"]
+			X, Y, Z = (ctrl.GetValue() for ctrl in ctrls)
+			XYZa = colormath.adapt(X, Y, Z, wp_src, wp_tgt, cat)
+			for i, ctrl in enumerate(ctrls):
+				ctrl.SetValue(XYZa[i])
+			self.parse_XYZ(color, False)
 	
 	def colorspace_ctrl_handler(self, event):
+		self.Freeze()
 		show = bool(self.colorspace_rgb_ctrl.Value)
 		for color in ("red", "green", "blue"):
 			getattr(self, "label_%s" % color).Show(show)
 			for component in "XYZxy":
 				getattr(self, "%s_%s" % (color, component)).Show(show)
-		self.enable_save_as_btn()
+		self.enable_btns()
 		self.update_layout()
+		self.Thaw()
 	
 	def drop_handler(self, path):
+		""" File dropped """
+		fn, ext = os.path.splitext(path)
+		if ext.lower() == ".ti3":
+			self.ti3_drop_handler(path)
+		else:
+			self.icc_drop_handler(path)
+
+	def icc_drop_handler(self, path):
+		""" ICC profile dropped """
 		try:
 			profile = ICCP.ICCProfile(path)
 		except (IOError, ICCP.ICCProfileInvalidError), exception:
 			show_result_dialog(Error(lang.getstr("profile.invalid") + "\n" +
 									 path), self)
 		else:
-			if profile.version >= 4:
+			if (profile.version >= 4 and
+				not profile.convert_iccv4_tags_to_iccv2()):
 				show_result_dialog(Error(lang.getstr("profile.iccv4.unsupported")),
 								   self)
 				return
@@ -289,49 +382,105 @@ class SynthICCFrame(BaseFrame):
 			except Exception, exception:
 				show_result_dialog(exception, self)
 			else:
-				self.panel.Freeze()
-				for ctrl, value in [(self.colorspace_rgb_ctrl,
-									 profile.colorSpace == "RGB"),
-									(self.colorspace_gray_ctrl,
-									 profile.colorSpace == "GRAY")]:
-					ctrl.SetValue(value)
-				self.colorspace_ctrl_handler(None)
 				if "lumi" in profile.tags:
 					luminance = profile.tags.lumi.Y
 				else:
 					luminance = 100
-				setcfg("synthprofile.luminance", luminance)
-				self.luminance_ctrl.SetValue(luminance)
-				for i, color in enumerate(("white", "black")):
-					for j, component in enumerate("XYZ"):
-						getattr(self, "%s_%s" %
-								(color, component)).SetValue(colors[i][j] /
-															 colors[0][1] * 100)
-					self.parse_XYZ(color)
-				for i, color in enumerate(("red", "green", "blue")):
-					xyY = colormath.XYZ2xyY(*colors[2 + i])
-					for j, component in enumerate("xy"):
-						getattr(self, "%s_%s" %
-								(color, component)).SetValue(xyY[j])
-				self.parse_xy(None)
-				self.black_XYZ_ctrl_handler(None)
-				trc = ICCP.CurveType(profile=profile)
-				for XYZ in colors[5:]:
-					trc.append(XYZ[1] / colors[0][1] * 65535)
-				transfer_function = trc.get_transfer_function()
-				if transfer_function and transfer_function[1] >= .95:
-					# Use detected transfer function
-					gamma = transfer_function[0][1]
+				if (not colors[1][1] and
+					isinstance(profile.tags.get("targ"), ICCP.Text)):
+					# The profile may not reflect the actual black point.
+					# Get it from the embedded TI3 instead if zero from lookup.
+					XYZbp = profile.get_chardata_bkpt(True)
+					if XYZbp:
+						# Use wtpt chromaticity
+						colors[1] = [v * XYZbp[1] for v in colors[0]]
+				self.set_colors(colors, luminance, profile.colorSpace)
+
+	def set_colors(self, colors, luminance, colorspace):
+		""" Set controls according to args """
+		self.panel.Freeze()
+		for ctrl, value in [(self.colorspace_rgb_ctrl,
+							 colorspace == "RGB"),
+							(self.colorspace_gray_ctrl,
+							 colorspace == "GRAY")]:
+			ctrl.SetValue(value)
+		self.colorspace_ctrl_handler(None)
+		setcfg("synthprofile.luminance", luminance)
+		self.luminance_ctrl.SetValue(luminance)
+		for i, color in enumerate(("white", "black")):
+			for j, component in enumerate("XYZ"):
+				getattr(self, "%s_%s" %
+						(color, component)).SetValue(colors[i][j] /
+													 colors[0][1] * 100)
+			self.parse_XYZ(color)
+		for i, color in enumerate(("red", "green", "blue")):
+			xyY = colormath.XYZ2xyY(*colors[2 + i])
+			for j, component in enumerate("xy"):
+				getattr(self, "%s_%s" %
+						(color, component)).SetValue(xyY[j])
+		self.parse_xy(None)
+		self.black_XYZ_ctrl_handler(None)
+		if len(colors[5:]) > 2:
+			trc = ICCP.CurveType()
+			for XYZ in colors[5:]:
+				trc.append(XYZ[1] / colors[0][1] * 65535)
+			transfer_function = trc.get_transfer_function(outoffset=1.0)
+			if transfer_function and transfer_function[1] >= .95:
+				# Use detected transfer function
+				gamma = transfer_function[0][1]
+			else:
+				# Use 50% gamma value
+				gamma = math.log(colors[132][1]) / math.log(128.0 / 255)
+			self.set_trc(round(gamma, 2))
+			setcfg("synthprofile.trc_gamma_type", "g")
+			self.trc_gamma_type_ctrl.SetSelection(self.trc_gamma_types_ba["g"])
+		self.panel.Thaw()
+
+	def ti3_drop_handler(self, path):
+		""" TI3 file dropped """
+		try:
+			ti3 = CGATS.CGATS(path)
+		except (IOError, CGATS.CGATSInvalidError), exception:
+			show_result_dialog(Error(lang.getstr("error.measurement.file_invalid",
+												 path)), self)
+		else:
+			ti3[0].normalize_to_y_100()
+			rgb = [(100, 100, 100), (0, 0, 0), (100, 0, 0), (0, 100, 0), (0, 0, 100)]
+			colors = []
+			for R, G, B in rgb:
+				result = ti3.queryi1({"RGB_R": R, "RGB_G": G, "RGB_B": B})
+				if result:
+					color = []
+					for component in "XYZ":
+						label = "XYZ_" + component
+						if label in result:
+							color.append(result[label])
+				if not result or len(color) < 3:
+					color = (0, 0, 0)
+				colors.append(color)
+			try:
+				(ti3_extracted,
+				 RGB_XYZ_extracted,
+				 RGB_XYZ_remaining) = extract_device_gray_primaries(ti3)
+			except Error, exception:
+				show_result_dialog(exception, self)
+			else:
+				RGB_XYZ_extracted.sort()
+				colors.extend(RGB_XYZ_extracted.values())
+				luminance = ti3.queryv1("LUMINANCE_XYZ_CDM2")
+				if luminance:
+					try:
+						luminance = float(luminance.split()[1])
+					except (TypeError, ValueError):
+						luminance = 100
 				else:
-					# Use 50% gamma value
-					gamma = math.log(colors[132][1]) / math.log(128.0 / 255)
-				self.set_trc(round(gamma, 2))
-				setcfg("synthprofile.trc_gamma_type", "g")
-				self.trc_gamma_type_ctrl.SetSelection(self.trc_gamma_types_ba["g"])
-				self.panel.Thaw()
+					luminance = 100
+				self.set_colors(colors, luminance, "RGB")
 	
-	def enable_save_as_btn(self):
-		self.save_as_btn.Enable(bool(self.get_XYZ()))
+	def enable_btns(self):
+		enable = bool(self.get_XYZ())
+		self.save_as_btn.Enable(enable)
+		self.chromatic_adaptation_btn.Enable(enable)
 	
 	def get_XYZ(self):
 		""" Get XYZ in 0..1 range """
@@ -363,10 +512,16 @@ class SynthICCFrame(BaseFrame):
 	def luminance_ctrl_handler(self, event):
 		v = self.luminance_ctrl.GetValue()
 		setcfg("synthprofile.luminance", v)
+		target_peak = v
+		maxmll = self.lut3d_hdr_maxmll_ctrl.GetValue()
+		if maxmll < target_peak:
+			setcfg("3dlut.hdr_maxmll", target_peak)
+		self.lut3d_hdr_maxmll_ctrl.SetRange(target_peak, 10000)
+		self.lut3d_set_option("3dlut.hdr_peak_luminance", v)
 		self.black_luminance_ctrl_handler(event)
 	
-	def parse_XYZ(self, name, set_blackpoint=False):
-		if not set_blackpoint:
+	def parse_XYZ(self, name, set_blackpoint=None):
+		if set_blackpoint is None:
 			set_blackpoint = not self.black_point_cb.Value
 		if not self._updating_ctrls:
 			self.preset_ctrl.SetSelection(0)
@@ -380,12 +535,18 @@ class SynthICCFrame(BaseFrame):
 			if name == "white" and set_blackpoint:
 				getattr(self, "black_%s" % (component)).SetValue(v * black_Y)
 		if "X" in XYZ and "Y" in XYZ and "Z" in XYZ:
-			xyY = colormath.XYZ2xyY(XYZ["X"], XYZ["Y"], XYZ["Z"])
+			if XYZ["X"] + XYZ["Y"] + XYZ["Z"] == 0:
+				# Set black chromaticity to white chromaticity if XYZ is 0
+				xyY = []
+				for i, component in enumerate("xy"):
+					xyY.append(getattr(self, "white_%s" % component).GetValue())
+			else:
+				xyY = colormath.XYZ2xyY(XYZ["X"], XYZ["Y"], XYZ["Z"])
 			for i, component in enumerate("xy"):
 				getattr(self, "%s_%s" % (name, component)).SetValue(xyY[i])
 				if name == "white" and set_blackpoint:
 					getattr(self, "black_%s" % (component)).SetValue(xyY[i])
-		self.enable_save_as_btn()
+		self.enable_btns()
 	
 	def parse_xy(self, name=None, set_blackpoint=False):
 		if not set_blackpoint:
@@ -438,11 +599,12 @@ class SynthICCFrame(BaseFrame):
 			for i, component in enumerate("XYZ"):
 				getattr(self, "%s_%s" %
 						(color, component)).SetValue(XYZ[color[0]][i] * 100)
-		self.enable_save_as_btn()
+		self.enable_btns()
 	
 	def preset_ctrl_handler(self, event):
 		preset_name = self.preset_ctrl.GetStringSelection()
 		if preset_name:
+			self.set_default_cat()
 			gamma, white, red, green, blue = colormath.rgb_spaces[preset_name]
 			white = colormath.get_whitepoint(white)
 			self._updating_ctrls = True
@@ -484,29 +646,35 @@ class SynthICCFrame(BaseFrame):
 			return "ok"
 		return "invalid"
 
+	def set_default_cat(self):
+		self.cat = "Bradford"
+
 	def set_trc(self, gamma):
 		if gamma == -1023:
 			# DICOM
 			self.trc_ctrl.SetSelection(1)
+		elif gamma == -2.0:
+			# HLG
+			self.trc_ctrl.SetSelection(2)
 		elif gamma == -3.0:
 			# L*
-			self.trc_ctrl.SetSelection(2)
+			self.trc_ctrl.SetSelection(3)
 		elif gamma == -709:
 			# Rec. 709
-			self.trc_ctrl.SetSelection(3)
+			self.trc_ctrl.SetSelection(4)
 		elif gamma == -1886:
 			# Rec. 1886
-			self.trc_ctrl.SetSelection(4)
+			self.trc_ctrl.SetSelection(5)
 			self.trc_ctrl_handler()
 		elif gamma == -240:
 			# SMPTE 240M
-			self.trc_ctrl.SetSelection(5)
+			self.trc_ctrl.SetSelection(6)
 		elif gamma == -2084:
 			# SMPTE 2084, roll-off clip
-			self.trc_ctrl.SetSelection(7)
+			self.trc_ctrl.SetSelection(8)
 		elif gamma == -2.4:
 			# sRGB
-			self.trc_ctrl.SetSelection(8)
+			self.trc_ctrl.SetSelection(9)
 		else:
 			# Gamma
 			self.trc_ctrl.SetSelection(0)
@@ -514,7 +682,7 @@ class SynthICCFrame(BaseFrame):
 		self.update_trc_controls()
 	
 	def profile_name_ctrl_handler(self, event):
-		self.enable_save_as_btn()
+		self.enable_btns()
 	
 	def red_XYZ_ctrl_handler(self, event):
 		self.parse_XYZ("red")
@@ -557,29 +725,32 @@ class SynthICCFrame(BaseFrame):
 			return
 
 		XYZ = self.get_XYZ()
-		if self.trc_ctrl.GetSelection() in (0, 4):
+		if self.trc_ctrl.GetSelection() in (0, 5):
 			# 0 = Gamma
-			# 4 = Rec. 1886 - trc set here is only used if black = 0
+			# 5 = Rec. 1886 - trc set here is only used if black = 0
 			trc = gamma
 		elif self.trc_ctrl.GetSelection() == 1:
 			# DICOM
 			trc = -1
 		elif self.trc_ctrl.GetSelection() == 2:
+			# HLG
+			trc = -2
+		elif self.trc_ctrl.GetSelection() == 3:
 			# L*
 			trc = -3.0
-		elif self.trc_ctrl.GetSelection() == 3:
+		elif self.trc_ctrl.GetSelection() == 4:
 			# Rec. 709
 			trc = -709
-		elif self.trc_ctrl.GetSelection() == 5:
+		elif self.trc_ctrl.GetSelection() == 6:
 			# SMPTE 240M
 			trc = -240
-		elif self.trc_ctrl.GetSelection() in (6, 7):
+		elif self.trc_ctrl.GetSelection() in (7, 8):
 			# SMPTE 2084
 			trc = -2084
-		elif self.trc_ctrl.GetSelection() == 8:
+		elif self.trc_ctrl.GetSelection() == 9:
 			# sRGB
 			trc = -2.4
-		rolloff = self.trc_ctrl.GetSelection() == 7
+		rolloff = self.trc_ctrl.GetSelection() == 8
 		class_i = self.profile_class_ctrl.GetSelection()
 		tech_i = self.tech_ctrl.GetSelection()
 		ciis_i = self.ciis_ctrl.GetSelection()
@@ -592,11 +763,11 @@ class SynthICCFrame(BaseFrame):
 				   "profile_class": self.profile_classes.keys()[class_i],
 				   "tech": self.tech.keys()[tech_i],
 				   "ciis": self.ciis.keys()[ciis_i]}
-		if trc == -2084 and rolloff and getcfg("synthprofile.luminance") < 10000:
-			if rolloff:
+		if (trc == -2084 and rolloff) or trc == -2:
+			if trc == -2084:
 				msg = "smpte2084.rolloffclip"
 			else:
-				msg = "smpte2084.hardclip"
+				msg = "hlg"
 			self.worker.recent.write(lang.getstr("trc." + msg) + "\n")
 			self.worker.start(consumer,
 							  self.create_profile, wargs=(XYZ, trc, path),
@@ -614,9 +785,11 @@ class SynthICCFrame(BaseFrame):
 											   (XYZ["gX"], XYZ["gY"], XYZ["gZ"]),
 											   (XYZ["bX"], XYZ["bY"], XYZ["bZ"]),
 											   (XYZ["wX"], XYZ["wY"], XYZ["wZ"]),
-											   trc,
+											   1.0,
 											   "",
-											   getcfg("copyright"))
+											   getcfg("copyright"),
+											   cat=self.cat,
+											   profile_class=profile_class)
 			black = colormath.adapt(XYZ["kX"], XYZ["kY"], XYZ["kZ"], white)
 			profile.tags.rTRC = ICCP.CurveType(profile=profile)
 			profile.tags.gTRC = ICCP.CurveType(profile=profile)
@@ -625,17 +798,26 @@ class SynthICCFrame(BaseFrame):
 		else:
 			# Grayscale profile
 			profile = ICCP.ICCProfile()
+			# Profile class
+			profile.profileClass = profile_class
+			if (not ICCP.s15f16_is_equal((XYZ["wX"], XYZ["wY"], XYZ["wZ"]),
+										 colormath.get_whitepoint("D50")) and
+				(profile.profileClass not in ("mntr", "prtr") or
+				 colormath.is_similar_matrix(colormath.get_cat_matrix(self.cat),
+										     colormath.get_cat_matrix("Bradford")))):
+				profile.version = 2.2  # Match ArgyllCMS
 			profile.colorSpace = "GRAY"
 			profile.setCopyright(getcfg("copyright"))
-			profile.tags.wtpt = ICCP.XYZType()
-			(profile.tags.wtpt.X,
-			 profile.tags.wtpt.Y,
-			 profile.tags.wtpt.Z) = (XYZ["wX"], XYZ["wY"], XYZ["wZ"])
+			profile.set_wtpt((XYZ["wX"], XYZ["wY"], XYZ["wZ"]), self.cat)
 			black = [XYZ["wY"] * (getcfg("synthprofile.black_luminance") /
 								  getcfg("synthprofile.luminance"))] * 3
 			profile.tags.kTRC = ICCP.CurveType(profile=profile)
 			channels = "k"
-		outoffset = getcfg("synthprofile.trc_output_offset")
+		if trc == -2:
+			# HLG
+			outoffset = 1
+		else:
+			outoffset = getcfg("synthprofile.trc_output_offset")
 		if trc == -1:
 			# DICOM
 			# Absolute luminance values!
@@ -661,19 +843,38 @@ class SynthICCFrame(BaseFrame):
 				# Grayscale profile
 				profile.tags.kTRC.set_bt1886_trc(black[1], outoffset, trc,
 												 getcfg("synthprofile.trc_gamma_type"))
-		elif trc == -2084:
-			# SMPTE 2084
+		elif trc == -2084 or trc == -2:
+			# SMPTE 2084 or HLG
+			if trc == -2084:
+				hdr_format = "PQ"
+			else:
+				hdr_format = "HLG"
+			minmll = getcfg("3dlut.hdr_minmll")
+			if rolloff:
+				maxmll = getcfg("3dlut.hdr_maxmll")
+			else:
+				maxmll = getcfg("synthprofile.luminance")
 			if rgb:
 				# Color profile
-				profile.set_smpte2084_trc([v * getcfg("synthprofile.luminance") *
-										   (1 - outoffset)
-										   for v in black],
-										  getcfg("synthprofile.luminance"),
-										  rolloff=rolloff,
-										  blend_blackpoint=False)
-				if rolloff and getcfg("synthprofile.luminance") < 10000:
+				if trc == -2084:
+					profile.set_smpte2084_trc([v * getcfg("synthprofile.luminance") *
+											   (1 - outoffset)
+											   for v in black],
+											  getcfg("synthprofile.luminance"),
+											  minmll, maxmll,
+											  getcfg("3dlut.hdr_maxmll_alt_clip"),
+											  rolloff=True,
+											  blend_blackpoint=False)
+				else:
+					# HLG
+					profile.set_hlg_trc((0, 0, 0),
+										getcfg("synthprofile.luminance"),
+										1.2,
+										getcfg("3dlut.hdr_ambient_luminance"),
+										blend_blackpoint=False)
+				if rolloff or trc == -2:
 					rgb_space = profile.get_rgb_space()
-					rgb_space[0] = -2084
+					rgb_space[0] = 1.0  # Set gamma to 1.0 (not actually used)
 					rgb_space = colormath.get_rgb_space(rgb_space)
 					linebuffered_logfiles = []
 					if sys.stdout.isatty():
@@ -687,24 +888,41 @@ class SynthICCFrame(BaseFrame):
 													   triggers=[])),
 									  self.worker.recent,
 									  self.worker.lastmsg])
-					profile.tags.A2B0 = ICCP.create_synthetic_smpte2084_clut_profile(
+					quality = getcfg("profile.quality")
+					clutres = {"m": 17, "l": 9}.get(quality, 33)
+					hdr_clut_profile = ICCP.create_synthetic_hdr_clut_profile(
+						hdr_format,
 						rgb_space, "",
 						getcfg("synthprofile.black_luminance") * (1 - outoffset),
-						getcfg("synthprofile.luminance"),
-						rolloff=rolloff, mode="ICtCp" if rolloff else "RGB",
-						worker=self.worker, logfile=logfiles).tags.A2B0
-					self.worker.generate_B2A_from_inverse_table(profile, 33,
-																smooth=False,
-																rgb_space=rgb_space,
-																logfile=logfiles)
-				if black != [0, 0, 0] and not bpc:
+						getcfg("synthprofile.luminance"), minmll, maxmll,
+						getcfg("3dlut.hdr_maxmll_alt_clip"),
+						1.2, getcfg("3dlut.hdr_ambient_luminance"),
+						clutres=clutres, sat=getcfg("3dlut.hdr_sat"),
+						hue=getcfg("3dlut.hdr_hue"),
+						generate_B2A=trc == -2, worker=self.worker,
+						logfile=logfiles,
+						cat=self.cat)
+					profile.tags.A2B0 = hdr_clut_profile.tags.A2B0
+					if trc == -2:
+						# HLG
+						profile.tags.B2A0 = hdr_clut_profile.tags.B2A0
+				if black != [0, 0, 0] and outoffset and not bpc:
 					profile.apply_black_offset(black)
 			else:
 				# Grayscale profile
-				profile.tags.kTRC.set_smpte2084_trc(getcfg("synthprofile.black_luminance") *
-													(1 - outoffset),
-													getcfg("synthprofile.luminance"),
-													rolloff=rolloff)
+				if trc == -2084:
+					profile.tags.kTRC.set_smpte2084_trc(getcfg("synthprofile.black_luminance") *
+														(1 - outoffset),
+														getcfg("synthprofile.luminance"),
+														minmll, maxmll,
+														getcfg("3dlut.hdr_maxmll_alt_clip"),
+														rolloff=True)
+				else:
+					# HLG
+					profile.tags.kTRC.set_hlg_trc(0,
+												  getcfg("synthprofile.luminance"),
+												  1.2,
+												  getcfg("3dlut.hdr_ambient_luminance"))
 				if black != [0, 0, 0] and outoffset and not bpc:
 					profile.tags.kTRC.apply_bpc(black[1])
 		elif black != [0, 0, 0]:
@@ -738,8 +956,6 @@ class SynthICCFrame(BaseFrame):
 			(profile.tags[tagname].X,
 			 profile.tags[tagname].Y,
 			 profile.tags[tagname].Z) = X, Y, Z
-		# Profile class
-		profile.profileClass = profile_class
 		# Technology type
 		if tech:
 			profile.tags.tech = ICCP.SignatureType("sig \0\0\0\0" +
@@ -794,7 +1010,7 @@ class SynthICCFrame(BaseFrame):
 		if not self._updating_ctrls:
 			self.preset_ctrl.SetSelection(0)
 		i = self.trc_ctrl.GetSelection()
-		if i == 4:
+		if i == 5:
 			# BT.1886
 			setcfg("synthprofile.trc_gamma", 2.4)
 			setcfg("synthprofile.trc_gamma_type", "G")
@@ -834,31 +1050,72 @@ class SynthICCFrame(BaseFrame):
 		self.update_trc_controls()
 
 	def update_trc_control(self):
-		if self.trc_ctrl.GetSelection() in (0, 4):
+		if self.trc_ctrl.GetSelection() in (0, 5):
 			if (getcfg("synthprofile.trc_gamma_type") == "G" and
 				getcfg("synthprofile.trc_output_offset") == 0 and
 				getcfg("synthprofile.trc_gamma") == 2.4):
-				self.trc_ctrl.SetSelection(4)  # BT.1886
+				self.trc_ctrl.SetSelection(5)  # BT.1886
 			else:
 				self.trc_ctrl.SetSelection(0)  # Gamma
 
 	def update_trc_controls(self):
 		i = self.trc_ctrl.GetSelection()
 		self.panel.Freeze()
-		self.trc_gamma_label.Show(i in (0, 4))
+		self.trc_gamma_label.Show(i in (0, 5))
 		self.trc_gamma_ctrl.SetValue(str(getcfg("synthprofile.trc_gamma")))
-		self.trc_gamma_ctrl.Show(i in (0, 4))
+		self.trc_gamma_ctrl.Show(i in (0, 5))
 		self.trc_gamma_type_ctrl.SetSelection(self.trc_gamma_types_ba[getcfg("synthprofile.trc_gamma_type")])
 		self.panel.GetSizer().Layout()
 		self.panel.Thaw()
-		if i in (0, 4, 6, 7):
+		if i in (0, 5, 7, 8):
+			# Gamma, BT.1886, SMPTE 2084 (PQ)
 			outoffset = int(getcfg("synthprofile.trc_output_offset") * 100)
 		else:
 			outoffset = 100
 		self.black_output_offset_ctrl.SetValue(outoffset)
 		self.black_output_offset_intctrl.SetValue(outoffset)
+		target_peak = getcfg("synthprofile.luminance")
+		maxmll = getcfg("3dlut.hdr_maxmll")
+		# Don't allow maxmll < target peak. Technically this restriction does
+		# not exist, but practically maxmll < target peak doesn't make sense.
+		if maxmll < target_peak:
+			maxmll = target_peak
+			setcfg("3dlut.hdr_maxmll", maxmll)
+		self.lut3d_hdr_maxmll_ctrl.SetRange(target_peak, 10000)
+		self.luminance_ctrl.SetValue(target_peak)
+		self.lut3d_hdr_minmll_ctrl.SetValue(getcfg("3dlut.hdr_minmll"))
+		self.lut3d_hdr_maxmll_ctrl.SetValue(maxmll)
+		self.lut3d_hdr_maxmll_alt_clip_cb.SetValue(not bool(getcfg("3dlut.hdr_maxmll_alt_clip")))
+		self.lut3d_hdr_sat_ctrl.SetValue(int(round(getcfg("3dlut.hdr_sat") * 100)))
+		self.lut3d_hdr_update_sat_val()
+		hue = int(round(getcfg("3dlut.hdr_sat") * 100))
+		self.lut3d_hdr_hue_ctrl.SetValue(hue)
+		self.lut3d_hdr_hue_intctrl.SetValue(hue)
+		setcfg("3dlut.hdr_peak_luminance", getcfg("synthprofile.luminance"))
+		self.lut3d_hdr_update_diffuse_white()
+		self.lut3d_hdr_ambient_luminance_ctrl.SetValue(getcfg("3dlut.hdr_ambient_luminance"))
+		self.lut3d_hdr_update_system_gamma()
+		self.lut3d_hdr_minmll_label.Show(i in (7, 8))  # SMPTE 2084 (PQ)
+		self.lut3d_hdr_minmll_ctrl.Show(i in (7, 8))  # SMPTE 2084 (PQ)
+		self.lut3d_hdr_minmll_ctrl_label.Show(i in (7, 8))  # SMPTE 2084 (PQ)
+		self.lut3d_hdr_maxmll_label.Show(i == 8)  # SMPTE 2084 (PQ)
+		self.lut3d_hdr_maxmll_ctrl.Show(i == 8)  # SMPTE 2084 (PQ)
+		self.lut3d_hdr_maxmll_ctrl_label.Show(i == 8)  # SMPTE 2084 (PQ)
+		self.lut3d_show_hdr_maxmll_alt_clip_ctrl()
+		self.lut3d_hdr_diffuse_white_label.Show(i == 8)  # SMPTE 2084 (PQ)
+		self.lut3d_hdr_diffuse_white_txt.Show(i == 8)  # SMPTE 2084 (PQ)
+		self.lut3d_hdr_diffuse_white_txt_label.Show(i == 8)  # SMPTE 2084 (PQ)
+		sizer = self.lut3d_hdr_sat_ctrl.ContainingSizer
+		sizer.ShowItems(i == 8)  # SMPTE 2084 (PQ)
+		sizer = self.lut3d_hdr_hue_ctrl.ContainingSizer
+		sizer.ShowItems(i == 8)  # SMPTE 2084 (PQ)
+		self.lut3d_hdr_ambient_luminance_label.Show(i == 2)  # HLG
+		self.lut3d_hdr_ambient_luminance_ctrl.Show(i == 2)  # HLG
+		self.lut3d_hdr_ambient_luminance_ctrl_label.Show(i == 2)  # HLG
+		self.lut3d_hdr_system_gamma_label.Show(i == 2)  # HLG
+		self.lut3d_hdr_system_gamma_txt.Show(i == 2)  # HLG
 		self.black_luminance_ctrl_handler(True)
-		if i in (3, 5):
+		if i in (4, 6):
 			# Rec 709 or SMPTE 240M
 			# Match Adobe 'video' profiles
 			self.profile_class_ctrl.SetStringSelection(self.profile_classes["scnr"])
